/******************************************************************************
 * @file     SpiGlue.c
 * @brief    M480 SPI Flash driver Sample file
 *
 * @note
 * Copyright (C) 2019 Nuvoton Technology Corp. All rights reserved.
 ******************************************************************************/

/*!<Includes */
#include <string.h>
#include "NuMicro.h"
#include "massstorage.h"
/*---------------------------------------------------------------------------*/
/* Define                                                                    */
/*---------------------------------------------------------------------------*/
#define SECTOR_NUM             512
#define FS_OK                  0
#define SPI_SECTOR_ERASE_SIZE  4*1024

#ifdef __ICCARM__
    #pragma data_alignment=32
    uint8_t SPI_BackupBuf[SPI_SECTOR_ERASE_SIZE];    // Backup Buffer to save SPI content
#else
    __align(32) uint8_t SPI_BackupBuf[SPI_SECTOR_ERASE_SIZE];    // Backup Buffer to save SPI content
#endif
/*---------------------------------------------------------------------------*/
/* Functions                                                                 */
/*---------------------------------------------------------------------------*/
static int SpiFlash_disk_read(uint32_t sector_no, int32_t number_of_sector, uint8_t *buff)
{
    //spiFlashRead(sector_no * SECTOR_NUM, number_of_sector * SECTOR_NUM, buff);
    SPIM_DMA_Read(sector_no * SECTOR_NUM, 0, number_of_sector * SECTOR_NUM, buff, CMD_DMA_FAST_READ, 1);
    return FS_OK;
}


static int SpiFlash_disk_write(uint32_t sector_no, int32_t number_of_sector, uint8_t *buff)
{
    int SPI_addr, SPI_AligneSectorAddr;
    int SPI_writeSize, SPI_tmpSize, SPI_TargetSize, SPI_RemainingSize, SPI_MaxWritetSize4Sector;
    uint32_t size_mask;

    size_mask = ~(SPI_SECTOR_ERASE_SIZE - 1);
    SPI_addr = sector_no * SECTOR_NUM;

    do {
        // Get the 4K alignment SPI flash address
        SPI_AligneSectorAddr = SPI_addr & size_mask;

        // Calculate the max available update SPI data size for current SPI address
        if (SPI_addr != SPI_AligneSectorAddr)
            SPI_MaxWritetSize4Sector = SPI_SECTOR_ERASE_SIZE - (SPI_addr - SPI_AligneSectorAddr);
        else
            SPI_MaxWritetSize4Sector = SPI_SECTOR_ERASE_SIZE;

        // Get the real write data size in this sector
        if (number_of_sector * SECTOR_NUM >= SPI_MaxWritetSize4Sector)
            SPI_writeSize = SPI_MaxWritetSize4Sector;
        else
            SPI_writeSize = number_of_sector * SECTOR_NUM;

        // Read the original Sector content
        SPIM_DMA_Read(SPI_AligneSectorAddr, 0, SPI_SECTOR_ERASE_SIZE, SPI_BackupBuf, CMD_DMA_FAST_READ, 1);

        // Erase the spicified Sector before write
        spiFlashEraseSector4Backup(SPI_AligneSectorAddr, 1); // Erase 1 sector size

        // Backup the data before Target address for backup
        SPI_tmpSize = 0;

        if (SPI_addr != SPI_AligneSectorAddr) {
            SPI_tmpSize = SPI_addr - SPI_AligneSectorAddr;
            SPIM_DMA_Write(SPI_AligneSectorAddr, 0, SPI_tmpSize, SPI_BackupBuf, CMD_NORMAL_PAGE_PROGRAM);
        }

        // Write the target SPI data
        SPI_TargetSize = SPI_writeSize;
        SPIM_DMA_Write(SPI_addr, 0, SPI_TargetSize, buff, CMD_NORMAL_PAGE_PROGRAM);

        SPI_addr = SPI_addr + SPI_TargetSize;
        buff += SPI_TargetSize;

        number_of_sector -= (SPI_TargetSize / 512);

        // Write the remaining data after target data if the size is less than on sector
        if (SPI_tmpSize + SPI_TargetSize < SPI_SECTOR_ERASE_SIZE) {
            SPI_RemainingSize = SPI_SECTOR_ERASE_SIZE - SPI_writeSize - SPI_tmpSize;
            SPIM_DMA_Write(SPI_addr, 0, SPI_RemainingSize, &SPI_BackupBuf[SPI_writeSize + SPI_tmpSize], CMD_NORMAL_PAGE_PROGRAM);
        }

    } while (number_of_sector > 0);

    return FS_OK;
}

int SPIM_Read(uint32_t uSector, uint32_t uBufcnt, uint8_t *uDAddr)
{
    int status = 0;
    status = SpiFlash_disk_read(uSector, uBufcnt, uDAddr);
    return status;
}

int SPIM_Write(uint32_t uSector, uint32_t uBufcnt, uint8_t *uSAddr)
{
    int status = 0;
    status = SpiFlash_disk_write(uSector, uBufcnt, uSAddr);
    return status;
}
int spiFlashEraseSector4Backup(uint32_t addr, uint32_t secCount)
{
    if ((addr % (SPI_SECTOR_ERASE_SIZE)) != 0)
        return -1;

    SPIM_EraseBlock(addr, 0, OPCODE_SE_4K, 1, 1);

    return Successful;
}

/*** (C) COPYRIGHT 2019 Nuvoton Technology Corp. ***/
